from AccessControl import getSecurityManager
from DateTime import DateTime
from plone.app.iterate.interfaces import IBaseline
from plone.app.iterate.interfaces import ICheckinCheckoutPolicy
from plone.app.iterate.interfaces import keys
from plone.app.iterate.permissions import CheckoutPermission
from plone.app.layout.viewlets.globalstatusmessage import MTYPES_DISPLAY
from plone.memoize.instance import memoize
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope.interface import implementer
from zope.viewlet.interfaces import IViewlet

import logging


logger = logging.getLogger("Plone")


@implementer(IViewlet)
class BaseInfoViewlet(BrowserView):
    def __init__(self, context, request, view, manager):
        super().__init__(context, request)
        self.__parent__ = view
        self.view = view
        self.manager = manager

    def update(self):
        pass

    def render(self):
        raise NotImplementedError

    @property
    @memoize
    def policy(self):
        return ICheckinCheckoutPolicy(self.context)

    @memoize
    def created(self):
        time = self.properties.get(keys.checkout_time, DateTime())
        util = getToolByName(self.context, "translation_service")
        return util.ulocalized_time(
            time,
            context=self.context,
            domain="plonelocales",
        )

    @memoize
    def creator(self):
        user_id = self.properties.get(keys.checkout_user)
        membership = getToolByName(self.context, "portal_membership")
        if not user_id:
            return membership.getAuthenticatedMember()
        return membership.getMemberById(user_id)

    @memoize
    def creator_url(self):
        creator = self.creator()
        if creator is not None:
            portal_url = getToolByName(self.context, "portal_url")
            return f"{portal_url()}/author/{creator.getId()}"

    @memoize
    def creator_name(self):
        creator = self.creator()
        if creator is not None:
            return creator.getProperty("fullname") or creator.getId()
        # User is not known by PAS. This may be due to LDAP issues, so we keep
        # the user and log this.
        name = self.properties.get(keys.checkout_user)
        if IBaseline.providedBy(self.context):
            warning_tpl = (
                "%s is a baseline of a plone.app.iterate checkout "
                'by an unknown user id "%s"'
            )
        else:
            # IWorkingCopy.providedBy(self.context)
            warning_tpl = (
                "%s is a working copy of a plone.app.iterate "
                'checkout by an unknown user id "%s"'
            )
        logger.warning(warning_tpl, self.context, name)
        return name

    @property
    @memoize
    def properties(self):
        ref = self._getReference()
        if ref:
            return self.policy.getProperties(ref, default={})
        else:
            return {}

    def _getReference(self):
        raise NotImplementedError

    def display_info_for_mtype(self, mtype):
        """get info for display of an mtype"""
        return MTYPES_DISPLAY.get(mtype, MTYPES_DISPLAY["info"])


class BaselineInfoViewlet(BaseInfoViewlet):
    index = ViewPageTemplateFile("info_baseline.pt")

    def render(self):
        sm = getSecurityManager()
        working_copy = self.working_copy()
        if working_copy is not None and (
            sm.checkPermission(ModifyPortalContent, self.context)
            or sm.checkPermission(CheckoutPermission, self.context)
            or sm.checkPermission(ModifyPortalContent, working_copy)
        ):
            return self.index()
        else:
            return ""

    @memoize
    def working_copy(self):
        return self.policy.getWorkingCopy()

    def _getReference(self):
        return self.working_copy()


class CheckoutInfoViewlet(BaseInfoViewlet):
    index = ViewPageTemplateFile("info_checkout.pt")

    def render(self):
        sm = getSecurityManager()
        baseline = self.baseline()
        if baseline is not None and (
            sm.checkPermission(ModifyPortalContent, self.context)
            or sm.checkPermission(CheckoutPermission, baseline)
        ):
            return self.index()
        else:
            return ""

    @memoize
    def baseline(self):
        return self.policy.getBaseline()

    def _getReference(self):
        return self.baseline()



from Acquisition import aq_inner
from Acquisition import aq_parent
from plone.app.iterate import interfaces
from plone.app.iterate.base import BaseContentCopier
from plone.app.iterate.dexterity import ITERATE_RELATION_NAME
from plone.app.iterate.dexterity.relation import StagingRelationValue
from plone.app.iterate.event import AfterCheckinEvent
from plone.app.relationfield.event import update_behavior_relations
from plone.dexterity.utils import createContentInContainer
from plone.dexterity.utils import iterSchemata
from Products.CMFCore.utils import getToolByName
from Products.DCWorkflow.DCWorkflow import DCWorkflowDefinition
from z3c.relationfield import event
from z3c.relationfield import RelationValue
from z3c.relationfield.event import updateRelations
from z3c.relationfield.schema import RelationChoice
from z3c.relationfield.schema import RelationList
from zc.relation.interfaces import ICatalog
from ZODB.PersistentMapping import PersistentMapping
from zope import component
from zope.annotation.interfaces import IAnnotations
from zope.event import notify
from zope.intid.interfaces import IIntIds
from zope.schema import getFieldsInOrder


class ContentCopier(BaseContentCopier):
    def copyTo(self, container):
        context = aq_inner(self.context)
        wc = self._copyBaseline(container)
        # get id of objects
        intids = component.getUtility(IIntIds)
        wc_id = intids.getId(wc)
        # create a relation
        relation = StagingRelationValue(wc_id)
        event._setRelation(context, ITERATE_RELATION_NAME, relation)
        return wc, relation

    def merge(self):
        baseline = self._getBaseline()

        # delete the working copy reference to the baseline
        self._deleteWorkingCopyRelation()
        # move the working copy to the baseline container, deleting the
        # baseline
        new_baseline = self._replaceBaseline(baseline)

        # patch the working copy with baseline info not preserved during
        # checkout
        self._reassembleWorkingCopy(new_baseline, baseline)

        return new_baseline

    def _replaceBaseline(self, baseline):
        wc_id = self.context.getId()
        wc_container = aq_parent(self.context)

        # copy all field values from the working copy to the baseline
        for schema in iterSchemata(baseline):
            for name, field in getFieldsInOrder(schema):
                # Skip read-only fields
                if field.readonly:
                    continue
                if field.__name__ == "id":
                    continue
                try:
                    value = field.get(schema(self.context))
                except Exception:
                    value = None

                if name == "effective":
                    baseline.effective_date = self.context.effective_date
                elif name == "expires":
                    baseline.expiration_date = self.context.expiration_date
                elif name == "subjects":
                    baseline.setSubject(self.context.Subject())
                else:
                    field.set(baseline, value)

        baseline.reindexObject()

        # copy annotations
        wc_annotations = IAnnotations(self.context)
        baseline_annotations = IAnnotations(baseline)

        # The next one is just wrong
        # baseline_annotations.clear()

        # Remove plone.folder.ordered.order from it to not mess with the original
        if "plone.folder.ordered.order" in wc_annotations:
            wc_annotations.pop("plone.folder.ordered.order")

        baseline_annotations.update(wc_annotations)

        # delete the working copy
        wc_container._delObject(wc_id)

        return baseline

    def _reassembleWorkingCopy(self, new_baseline, baseline):
        # reattach the source's workflow history, try avoid a dangling ref
        try:
            new_baseline.workflow_history = PersistentMapping(
                baseline.workflow_history.items()
            )
        except AttributeError:
            # No workflow apparently.  Oh well.
            pass

        # reset wf state security directly
        workflow_tool = getToolByName(self.context, "portal_workflow")
        wfs = workflow_tool.getWorkflowsFor(self.context)
        for wf in wfs:
            if not isinstance(wf, DCWorkflowDefinition):
                continue
            wf.updateRoleMappingsFor(new_baseline)
        return new_baseline

    def _handleReferences(self, baseline, wc, mode, wc_ref):
        pass

    def _deleteWorkingCopyRelation(self):
        # delete the wc reference keeping a reference to it for its annotations
        relation = self._get_relation_to_baseline()
        catalog = component.queryUtility(ICatalog)
        catalog.unindex(relation)
        return

    def _get_relation_to_baseline(self):
        context = aq_inner(self.context)
        # get id
        intids = component.getUtility(IIntIds)
        id = intids.getId(context)
        # ask catalog
        catalog = component.getUtility(ICatalog)
        relations = catalog.findRelations({"to_id": id})
        relations = [i for i in relations if i.from_attribute == ITERATE_RELATION_NAME]
        # do we have a baseline in our relations?
        if relations and not len(relations) == 1:
            raise interfaces.CheckinException("Baseline count mismatch")

        if not relations or not relations[0]:
            raise interfaces.CheckinException("Baseline has disappeared")
        return relations[0]

    def _getBaseline(self):
        intids = component.getUtility(IIntIds)
        relation = self._get_relation_to_baseline()
        if relation:
            baseline = intids.getObject(relation.from_id)

        if not baseline:
            raise interfaces.CheckinException("Baseline has disappeared")
        return baseline

    def checkin(self, checkin_message):
        # get the baseline for this working copy, raise if not found
        baseline = self._getBaseline()
        # get a hold of the relation object
        relation = self._get_relation_to_baseline()
        # publish the event for subscribers, early because contexts are about
        # to be manipulated
        notify(event.CheckinEvent(self.context, baseline, relation, checkin_message))
        # merge the object back to the baseline with a copier
        copier = component.queryAdapter(self.context, interfaces.IObjectCopier)
        new_baseline = copier.merge()
        # don't need to unlock the lock disappears with old baseline deletion
        notify(AfterCheckinEvent(new_baseline, checkin_message))
        return new_baseline


class FolderishContentCopier(ContentCopier):
    def _copyBaseline(self, container):
        obj = createContentInContainer(
            container,
            self.context.portal_type,
            id=f"working_copy_of_{self.context.id}",
        )

        # copy all field values from the baseline to the working copy
        for schema in iterSchemata(self.context):
            for name, field in getFieldsInOrder(schema):
                # Skip read-only fields
                if field.readonly:
                    continue
                if field.__name__ == "id":
                    continue
                try:
                    value = field.get(schema(self.context))
                except Exception:
                    value = None

                if name == "effective":
                    obj.effective_date = self.context.effective_date
                elif name == "expires":
                    obj.expiration_date = self.context.expiration_date
                elif name == "subjects":
                    obj.setSubject(self.context.Subject())
                elif isinstance(field, RelationList):
                    if value:
                        field.set(
                            obj,
                            [RelationValue(i.to_id) for i in value if not i.isBroken()],
                        )
                elif isinstance(field, RelationChoice):
                    if value and not value.isBroken():
                        field.set(obj, RelationValue(value.to_id))
                else:
                    field.set(obj, value)

        update_relation_catalog(obj)
        obj.reindexObject()

        # copy annotations
        wc_annotations = IAnnotations(obj)
        baseline_annotations = IAnnotations(self.context)

        wc_annotations.update(baseline_annotations)
        # Remove plone.folder.order from it to not spoil the wc
        if "plone.folder.ordered.order" in wc_annotations:
            wc_annotations.pop("plone.folder.ordered.order")
        if "plone.folder.ordered.pos" in wc_annotations:
            wc_annotations.pop("plone.folder.ordered.pos")

        return obj

    def _replaceBaseline(self, baseline):
        wc_id = self.context.getId()
        wc_container = aq_parent(self.context)

        # copy all field values from the working copy to the baseline
        for schema in iterSchemata(baseline):
            for name, field in getFieldsInOrder(schema):
                # Skip read-only fields
                if field.readonly:
                    continue
                if field.__name__ == "id":
                    continue
                try:
                    value = field.get(schema(self.context))
                except Exception:
                    value = None

                if name == "effective":
                    baseline.effective_date = self.context.effective_date
                elif name == "expires":
                    baseline.expiration_date = self.context.expiration_date
                elif name == "subjects":
                    baseline.setSubject(self.context.Subject())
                elif isinstance(field, RelationList):
                    if value:
                        field.set(
                            baseline,
                            [RelationValue(i.to_id) for i in value if not i.isBroken()],
                        )
                elif isinstance(field, RelationChoice):
                    if value and not value.isBroken():
                        field.set(baseline, RelationValue(value.to_id))
                else:
                    field.set(baseline, value)
        update_relation_catalog(baseline)
        baseline.reindexObject()

        # Move working children (newly created objects)
        working_copy_children = [a for a in self.context.objectIds()]

        clipboard = self.context.manage_cutObjects(working_copy_children)
        baseline.manage_pasteObjects(clipboard)

        # copy back annotations
        wc_annotations = IAnnotations(self.context)
        baseline_annotations = IAnnotations(baseline)

        # Remove plone.folder.ordered.order from it to not mess with the original
        if "plone.folder.ordered.order" in wc_annotations:
            wc_annotations.pop("plone.folder.ordered.order")
        if "plone.folder.ordered.pos" in wc_annotations:
            wc_annotations.pop("plone.folder.ordered.pos")

        baseline_annotations.update(wc_annotations)

        # delete the working copy
        wc_container._delObject(wc_id)

        return baseline


def update_relation_catalog(obj):
    # updateRelations from z3c.relationfield does not properly update relations in behaviors
    # that are registered with a marker-interface.
    # update_behavior_relations from plone.app.relationfield does that but does not update
    # those in the main schema.
    updateRelations(obj, None)
    update_behavior_relations(obj, None)



##################################################################
#
# (C) Copyright 2006 ObjectRealms, LLC
# All Rights Reserved
#
# This file is part of iterate.
#
# iterate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# iterate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iterate; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################

from DateTime import DateTime
from plone.app.iterate.browser.control import Control
from plone.app.iterate.interfaces import ICheckinCheckoutPolicy
from plone.app.iterate.testing import PLONEAPPITERATEDEX_INTEGRATION_TESTING
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from zc.relation.interfaces import ICatalog
from zope import component
from zope.intid.interfaces import IIntIds

import unittest


class TestIterations(unittest.TestCase):
    layer = PLONEAPPITERATEDEX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        login(self.portal, TEST_USER_NAME)

        self.wf = self.portal.portal_workflow
        self.wf.setChainForPortalTypes(("Document",), "plone_workflow")

        # add a folder with two documents in it
        self.portal.invokeFactory("Folder", "docs")
        self.portal.docs.invokeFactory("Document", "doc1")
        self.portal.docs.invokeFactory("Document", "doc2")
        self.portal.docs.invokeFactory("FolderishDocument", "doc3")

        # add a working copy folder
        self.portal.invokeFactory("Folder", "workarea")

        self.repo = self.portal.portal_repository

    def test_baselineVersionCreated(self):
        # if a baseline has no version ensure that one is created on checkout

        doc = self.portal.docs.doc1
        self.assertTrue(self.repo.isVersionable(doc))

        history = self.repo.getHistory(doc)
        self.assertEqual(len(history), 0)

        ICheckinCheckoutPolicy(doc).checkout(self.portal.workarea)

        history = self.repo.getHistory(doc)
        self.assertEqual(len(history), 1)

        doc2 = self.portal.docs.doc2
        self.repo.save(doc2)

        ICheckinCheckoutPolicy(doc2).checkout(self.portal.workarea)

        history = self.repo.getHistory(doc2)
        self.assertEqual(len(history), 1)

    def test_folderOrder(self):
        """When an item is checked out and then back in, the original
        folder order is preserved."""
        container = self.portal.docs
        doc = container.doc1
        original_position = container.getObjectPosition(doc.getId())

        # check that there is another document which could interact with
        # position of document the test work on
        doc2_position = container.getObjectPosition("doc2")
        self.assertTrue(doc2_position > original_position)

        self.repo.save(doc)
        wc = ICheckinCheckoutPolicy(doc).checkout(container)
        wc.text = "new document text"

        # check that the copy is put after the second document
        copy_position = container.getObjectPosition(wc.getId())
        self.assertTrue(copy_position > doc2_position)

        new_doc = ICheckinCheckoutPolicy(wc).checkin("updated")
        new_position = container.getObjectPosition(new_doc.getId())
        self.assertEqual(new_position, original_position)

    def test_default_page_is_kept_in_folder(self):
        # Ensure that a default page that is checked out and back in is still
        # the default page.
        folder = self.portal.docs
        doc = folder.doc1
        from Products.CMFDynamicViewFTI.interfaces import ISelectableBrowserDefault

        ISelectableBrowserDefault(folder).setDefaultPage("doc1")
        self.assertEqual(folder.getProperty("default_page", ""), "doc1")
        self.assertEqual(folder.getDefaultPage(), "doc1")
        # Note: when checking out to self.portal.workarea it surprisingly works
        # without changes.  But the default behavior in Plone is to check a
        # document out in its original folder, so that is what we check here.
        wc = ICheckinCheckoutPolicy(doc).checkout(folder)
        doc = ICheckinCheckoutPolicy(wc).checkin("updated")
        self.assertEqual(folder.getProperty("default_page", ""), "doc1")
        self.assertEqual(folder.getDefaultPage(), "doc1")

    def test_control_checkin_allowed_with_no_policy(self):
        control = Control(self.portal, self.layer["request"])
        self.assertFalse(control.checkin_allowed())

    def test_control_checkout_allowed_with_no_policy(self):
        control = Control(self.portal, self.layer["request"])
        self.assertFalse(control.checkout_allowed())

    def test_control_cancel_allowed_with_no_policy(self):
        control = Control(self.portal, self.layer["request"])
        self.assertFalse(control.cancel_allowed())

    def test_control_cancel_on_original_does_not_delete_original(self):
        # checkout document
        doc = self.portal.docs.doc1
        policy = ICheckinCheckoutPolicy(self.portal.docs.doc1, None)
        policy.checkout(self.portal.workarea)

        # get cancel browser view
        from plone.app.iterate.browser.cancel import Cancel

        cancel = Cancel(doc, self.layer["request"])
        self.layer["request"].form["form.button.Cancel"] = True

        # check if cancel on original raises the correct exception
        from plone.app.iterate.interfaces import CheckoutException

        with self.assertRaises(CheckoutException):
            cancel()

    def test_relationship_deleted_on_checkin(self):
        # Ensure that the relationship between the baseline and the wc is
        # removed when the working copy is checked in
        folder = self.portal.docs
        doc = folder.doc1

        intids = component.getUtility(IIntIds)
        catalog = component.getUtility(ICatalog)
        obj_id = intids.getId(doc)
        rels = list(catalog.findRelations({"from_id": obj_id}))
        self.assertEqual(len(rels), 0)

        wc = ICheckinCheckoutPolicy(doc).checkout(folder)
        wc_id = intids.getId(wc)

        rels = list(catalog.findRelations({"from_id": obj_id}))
        self.assertEqual(len(rels), 1)
        from_rel = rels[0]

        rels = list(catalog.findRelations({"to_id": wc_id}))
        self.assertEqual(len(rels), 1)
        to_rel = rels[0]

        self.assertEqual(from_rel, to_rel)

        doc = ICheckinCheckoutPolicy(wc).checkin("updated")

        rels = list(catalog.findRelations({"from_id": obj_id}))

        self.assertEqual(len(rels), 0)

    def test_relationship_deleted_on_cancel_checkout(self):
        # Ensure that the relationship between the baseline and the wc is
        # removed when the working copy is checked in
        folder = self.portal.docs
        doc = folder.doc1

        intids = component.getUtility(IIntIds)
        catalog = component.getUtility(ICatalog)
        obj_id = intids.getId(doc)
        rels = list(catalog.findRelations({"from_id": obj_id}))
        self.assertEqual(len(rels), 0)

        wc = ICheckinCheckoutPolicy(doc).checkout(folder)
        wc_id = intids.getId(wc)

        rels = list(catalog.findRelations({"from_id": obj_id}))
        self.assertEqual(len(rels), 1)
        from_rel = rels[0]

        rels = list(catalog.findRelations({"to_id": wc_id}))
        self.assertEqual(len(rels), 1)
        to_rel = rels[0]

        self.assertEqual(from_rel, to_rel)

        ICheckinCheckoutPolicy(wc).cancelCheckout()

        rels = list(catalog.findRelations({"from_id": obj_id}))

        self.assertEqual(len(rels), 0)

    def test_baseline_relations_updated_on_checkin(self):
        # Ensure that relations between the baseline and
        # and other objects are up-to-date on checkin
        from z3c.relationfield import RelationValue
        from zope.event import notify
        from zope.lifecycleevent import ObjectModifiedEvent

        folder = self.portal.docs
        baseline = folder.doc1
        target = folder.doc2

        intids = component.getUtility(IIntIds)
        catalog = component.getUtility(ICatalog)

        target_id = intids.getId(target)
        target_rel = [RelationValue(target_id)]

        # Test, if nothing is present in the relation catalog
        rels = list(catalog.findRelations({"to_id": target_id}))
        self.assertEqual(len(rels), 0)

        # set relatedItems on baseline
        baseline.relatedItems = target_rel
        notify(ObjectModifiedEvent(baseline))

        # Test, if relation is present in the relation catalog
        rels = list(catalog.findRelations({"to_id": target_id}))
        self.assertEqual(len(rels), 1)

        # make a workingcopy from baseline
        wc = ICheckinCheckoutPolicy(baseline).checkout(folder)

        # remove relations on wc
        wc.relatedItems = []
        notify(ObjectModifiedEvent(wc))

        # baseline -> target relation should be still there, we are on wc
        rels = list(catalog.findRelations({"to_id": target_id}))
        self.assertEqual(len(rels), 1)

        # baseline -> target relation should be empty now
        # because we replaced our baseline with our wc (with empty relations)
        baseline = ICheckinCheckoutPolicy(wc).checkin("updated")
        rels = list(catalog.findRelations({"to_id": target_id}))

        # new baseline's relatedItems should be empty
        self.assertEqual(len(rels), 0)

    def test_publication_behavior_values_not_changed(self):
        doc = self.portal.docs.doc1
        original_effective_date = doc.effective_date
        original_expiration_date = doc.expiration_date
        # Create a working copy
        wc = ICheckinCheckoutPolicy(doc).checkout(self.portal.workarea)
        # Check in without modifying the existing values
        baseline = ICheckinCheckoutPolicy(wc).checkin("updated")
        # Values should be the same of the original document
        self.assertEqual(original_effective_date, baseline.effective_date)
        self.assertEqual(original_expiration_date, baseline.expiration_date)

    def test_publication_behavior_values_not_changed_value_is_already_set(self):
        doc = self.portal.docs.doc2
        effective_date = DateTime("2021/09/08 10:06:00 UTC")
        doc.effective_date = effective_date
        original_expiration_date = doc.expiration_date
        # Create a working copy
        wc = ICheckinCheckoutPolicy(doc).checkout(self.portal.workarea)
        # Check in without modifying the existing values
        baseline = ICheckinCheckoutPolicy(wc).checkin("updated")
        # Values should be the same of the original document
        self.assertEqual(effective_date, baseline.effective_date)
        self.assertEqual(original_expiration_date, baseline.expiration_date)

    def test_publication_behavior_values_not_changed_folderish(self):
        doc = self.portal.docs.doc3
        original_effective_date = doc.effective_date
        original_expiration_date = doc.expiration_date
        # Create a working copy
        wc = ICheckinCheckoutPolicy(doc).checkout(self.portal.workarea)
        # Check in without modifying the existing values
        baseline = ICheckinCheckoutPolicy(wc).checkin("updated")
        # Values should be the same of the original document
        self.assertEqual(original_effective_date, baseline.effective_date)
        self.assertEqual(original_expiration_date, baseline.expiration_date)



##################################################################
#
# (C) Copyright 2006 ObjectRealms, LLC
# All Rights Reserved
#
# This file is part of iterate.
#
# iterate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# iterate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iterate; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################

from AccessControl import getSecurityManager
from Acquisition import aq_inner
from plone.app.iterate import interfaces
from plone.app.iterate.interfaces import ICheckinCheckoutPolicy
from plone.app.iterate.interfaces import IWorkingCopy
from plone.app.iterate.permissions import CheckinPermission
from plone.app.iterate.permissions import CheckoutPermission
from plone.memoize.view import memoize
from Products.Five.browser import BrowserView

import Products.CMFCore.permissions


class Control(BrowserView):
    """Information about whether iterate can operate.

    This is a public view, referenced in action condition expressions.
    """

    def checkin_allowed(self):
        """Check if a checkin is allowed"""
        context = aq_inner(self.context)
        checkPermission = getSecurityManager().checkPermission

        if not interfaces.IIterateAware.providedBy(context):
            return False

        archiver = interfaces.IObjectArchiver(context)
        if not archiver.isVersionable():
            return False

        if not IWorkingCopy.providedBy(context):
            return False

        policy = ICheckinCheckoutPolicy(context, None)
        if policy is None:
            return False

        original = policy.getBaseline()
        if original is None:
            return False

        can_modify = checkPermission(
            Products.CMFCore.permissions.ModifyPortalContent,
            original,
        )
        if not can_modify:
            return False

        can_checkin = checkPermission(
            CheckinPermission,
            original,
        )
        if not can_checkin:
            return False

        return True

    def checkout_allowed(self):
        """Check if a checkout is allowed."""
        context = aq_inner(self.context)

        if not interfaces.IIterateAware.providedBy(context):
            return False

        archiver = interfaces.IObjectArchiver(context)
        if not archiver.isVersionable():
            return False

        policy = ICheckinCheckoutPolicy(context, None)
        if policy is None:
            return False

        if policy.getWorkingCopy() is not None:
            return False

        # check if its is a checkout
        if policy.getBaseline() is not None:
            return False

        checkPermission = getSecurityManager().checkPermission
        can_checkout = checkPermission(
            CheckoutPermission,
            context,
        )
        if not can_checkout:
            return False

        return True

    @memoize
    def cancel_allowed(self):
        """Check to see if the user can cancel the checkout on the
        given working copy
        """
        policy = ICheckinCheckoutPolicy(self.context, None)
        if policy is None:
            return False
        original = policy.getBaseline()
        return original is not None



from plone.app.iterate.dexterity.interfaces import IStagingRelationValue
from z3c.relationfield import relation
from zc.relation.interfaces import ICatalog
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.component import getUtility
from zope.interface import implementer
from zope.intid.interfaces import IIntIds


@implementer(IStagingRelationValue, IAttributeAnnotatable)
class StagingRelationValue(relation.RelationValue):
    @classmethod
    def get_relations_of(cls, obj, from_attribute=None):
        """a list of relations to or from the passed object"""
        catalog = getUtility(ICatalog)
        intids = getUtility(IIntIds)
        obj_id = intids.getId(obj)
        items = list(catalog.findRelations({"from_id": obj_id}))
        items += list(catalog.findRelations({"to_id": obj_id}))
        if from_attribute:

            def condition(r):
                return r.from_attribute == from_attribute and not r.is_broken()

            items = filter(condition, items)
        return items



from plone.app.iterate.interfaces import IBaseline
from plone.app.iterate.interfaces import ICheckinCheckoutPolicy
from plone.app.iterate.interfaces import IWorkingCopy
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView


class DiffView(BrowserView):
    def __call__(self):
        policy = ICheckinCheckoutPolicy(self.context)
        if IBaseline.providedBy(self.context):
            self.baseline = self.context
            self.working_copy = policy.getWorkingCopy()
        elif IWorkingCopy.providedBy(self.context):
            self.working_copy = self.context
            self.baseline = policy.getBaseline()
        else:
            raise AttributeError("Invalid Context")
        return self.index()

    def diffs(self):
        diff = getToolByName(self.context, "portal_diff")
        return diff.createChangeSet(
            self.baseline, self.working_copy, id1="Baseline", id2="Working Copy"
        )



##################################################################
#
# (C) Copyright 2006-2007 ObjectRealms, LLC
# All Rights Reserved
#
# This file is part of iterate.
#
# iterate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# iterate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CMFDeployment; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################
from . import interfaces
from Products.CMFCore.utils import getToolByName
from zope.component import adapter
from zope.interface import implementer


@implementer(interfaces.IObjectArchiver)
@adapter(interfaces.IIterateAware)
class ContentArchiver:
    def __init__(self, context):
        self.context = context
        self.repository = getToolByName(context, "portal_repository")

    def save(self, checkin_message):
        self.repository.save(self.context, checkin_message)

    def isVersionable(self):
        if not self.repository.isVersionable(self.context):
            return False
        return True

    def isVersioned(self):
        archivist = getToolByName(self.context, "portal_archivist")
        version_count = len(archivist.queryHistory(self.context))
        return bool(version_count)

    def isModified(self):
        try:
            return not self.repository.isUpToDate(self.context)
        except Exception:
            return False
